/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamechobase.h"
#include <gdam/gdamsample.h>

/* 
 *    ____ _                 ___       _ _   
 *   / ___| | __ _ ___ ___  |_ _|_ __ (_) |_ 
 *  | |   | |/ _` / __/ __|  | || '_ \| | __|
 *  | |___| | (_| \__ \__ \  | || | | | | |_ 
 *   \____|_|\__,_|___/___/ |___|_| |_|_|\__|
 */

/* Functions for echo base class */
static gboolean gdam_echo_base_real_refresh_echo_array(GdamEchoBase* echo_base);
/* Functions for filter class */
/*
static void echo_base_on_child_set_source(GdamFilter* filter,
					  guint new_source_id);
*/
static void gdam_echo_base_on_child_set_beat_info(GdamFilter*           filter,
					     const GdamBeatInfo*   beat_info);
static int gdam_echo_base_append_start_request(GdamFilter* filter,
				               int new_source_id,
				               GdamRequestBuilder* builder);

static void gdam_echo_base_arg_set_reverb(GtkObject* object, GdamArg* arg)
{
	GdamEchoBase* echo_base = GDAM_ECHO_BASE(object);
	echo_base->do_reverb = GDAM_ARG_INT(arg);
	gdam_echo_base_refresh_echo_array(echo_base);
}

static void gdam_echo_base_arg_get_reverb(GtkObject* object, GdamArg* arg)
{
	GdamEchoBase* echo_base = GDAM_ECHO_BASE(object);
	GDAM_ARG_INT(arg) = echo_base->do_reverb;
}

static void gdam_echo_base_class_init(GdamEchoBaseClass* klass)
{
	GdamFilterClass* filter_class = GDAM_FILTER_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	klass->refresh_echo_array
				= gdam_echo_base_real_refresh_echo_array;
	filter_class->on_child_set_beat_info
				= gdam_echo_base_on_child_set_beat_info;
	filter_class->append_start_request
				= gdam_echo_base_append_start_request;
	gdam_arg_register(object_class,
			  "reverb",
			  GDAM_ARG_TYPE_INT,
			  gdam_echo_base_arg_set_reverb,
			  gdam_echo_base_arg_get_reverb,
			  GDAM_ARG_VITAL);
}

static void gdam_echo_base_init(GdamEchoBase* echo)
{
	echo->num_echoes = 1;
	echo->echoes = g_renew(int, echo->echoes, echo->num_echoes*2);
	echo->echoes[0] = (int)GdamFixed_ONE;
	echo->echoes[1] = 0;
} 

GtkType gdam_echo_base_get_type()
{
	static GtkType echo_base_type = 0;
	if (!echo_base_type) {
		static const GtkTypeInfo echo_info = {
			"GdamEchoBase",
			sizeof(GdamEchoBase),
			sizeof(GdamEchoBaseClass),
			(GtkClassInitFunc) gdam_echo_base_class_init,
			(GtkObjectInitFunc) gdam_echo_base_init,
			/* reserved 1 */ NULL,
			/* reserved 2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		echo_base_type = gtk_type_unique(GDAM_TYPE_FILTER, &echo_info);
	}
	return echo_base_type;
}

/*
 *     ____      _ _ _                _        
 *    / ___|__ _| | | |__   __ _  ___| | _____ 
 *   | |   / _` | | | '_ \ / _` |/ __| |/ / __|
 *   | |__| (_| | | | |_) | (_| | (__|   <\__ \
 *    \____\__,_|_|_|_.__/ \__,_|\___|_|\_\___/
 */

static void gdam_echo_base_on_child_set_beat_info(GdamFilter*           filter,
			                     const GdamBeatInfo*   beat_info)
{
	gdam_model_set_beat_info(GDAM_MODEL(filter), beat_info);
	if (!gdam_echo_base_refresh_echo_array(GDAM_ECHO_BASE(filter)))
		g_warning("error refreshing echo array");
}

static int gdam_echo_base_append_start_request(GdamFilter* filter,
				               int new_source_id,
				               GdamRequestBuilder* builder)
{
	GdamEchoBase* echo_base = GDAM_ECHO_BASE(filter);
	GdamRequest* request;
	gtk_object_ref(GTK_OBJECT(filter));
	GDAM_MODEL(filter)->create_source_pending = 1;
	request = gdam_request_create_source("GdamEcho",
					     gdam_model_got_subsource,
					     gdam_model_didnt_get_subsource,
					     echo_base,
					     2,
					     "filtered",
					     GDAM_ARG_TYPE_INT,
					     new_source_id,
					     "echoes",
					     GDAM_ARG_TYPE_INT_ARRAY,
					     echo_base->num_echoes * 2,
					     echo_base->echoes);
	return gdam_request_builder_append_create(builder, request);
}

static gboolean check_echo_array(GdamEchoBase* echo_base)
{
	int i;
	int last_time = 0;

	for (i=1; i<echo_base->num_echoes*2; i+=2){
		if (echo_base->echoes[i] < 0){
			g_warning("negative time in echo array!!");
			return FALSE;
		}
#if 0
		/* A good idea, but no longer required by the server. */
		if (echo_base->echoes[i] < last_time){
			g_warning("de-sequencified times in echo array!!");
			return FALSE;
		}
#endif
		last_time = echo_base->echoes[i];
	}
	return TRUE;
}

static void set_arg_complete(gpointer data)
{
	GdamEchoBase* echo_base = GDAM_ECHO_BASE(data);
	if (echo_base->pending_request != NULL)
	{
		gdam_debug("handling pending_request..");
		gdam_channel_append_request(GDAM_MODEL(echo_base)->channel, 
					    echo_base->pending_request);
		echo_base->pending_request = NULL;
	}
	else {
		echo_base->making_request = FALSE;
	}
}

void gdam_echo_base_update(GdamEchoBase*      echo_base)
{
	GdamModel* model = GDAM_MODEL(echo_base);
	GdamRequest* request;
	GdamArg args[2];

	if (model->create_source_pending){
		g_warning("Error, echo array lost while create pending..");
		return;		/* TODO: should handle `pending create' */
	}
	if (model->source_id == 0)
		return;

	if (!check_echo_array(echo_base)){
		g_warning("Error, bad echo array.");
		return;
	}

	args[0].name = "reverb";
	args[0].type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT(args+0) = echo_base->do_reverb;

	args[1].name = "echoes";
	args[1].type = GDAM_ARG_TYPE_INT_ARRAY;
	GDAM_ARG_INT_ARRAY(args+1).length = echo_base->num_echoes * 2;
	GDAM_ARG_INT_ARRAY(args+1).array = echo_base->echoes;
	request = gdam_request_set_args(
			model->source_id,
			2,
			args,
			set_arg_complete, 
			set_arg_complete, 
			model);
	if (echo_base->making_request){
		if (echo_base->pending_request != NULL)
			gdam_request_destroy(echo_base->pending_request);
		echo_base->pending_request = request;
	}
	else {
		echo_base->making_request = TRUE;
		gdam_channel_append_request(model->channel, request);
	}
}

gboolean gdam_echo_base_refresh_echo_array(GdamEchoBase* echo_base)
{
	GdamEchoBaseClass* klass;
	klass = GDAM_ECHO_BASE_CLASS(GTK_OBJECT(echo_base)->klass);
	if ( ! (*klass->refresh_echo_array)(echo_base) )
		return FALSE;
	gdam_echo_base_update(echo_base);
	return TRUE;
}

static gboolean gdam_echo_base_real_refresh_echo_array(GdamEchoBase* echo_base)
{
	g_warning("echo_base refresh_echo_array should not be called.");
	return FALSE;
}

