/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamblockfilter.h"
#include "gdamdispatch.h"
#include "gdamnamespace.h"


static gboolean gdam_block_filter_is_ready_to_read (GdamSource*      source);
static int      gdam_block_filter_read             (GdamSource*	     filter,
                                       	            GdamSample*	     buf,
				       	            int		     n);
static void     gdam_block_filter_miss_data        (GdamSource*      input,
					            int              lost);
static int 	gdam_block_filter_process_block    (GdamBlockFilter* filter,
	                                            GdamSample*      output,
					            const GdamSample*input);

static void     gdam_block_filter_class_init       (GdamBlockFilterClass*);
static void     gdam_block_filter_init             (GdamBlockFilter* filter);
static void     gdam_block_filter_destroy          (GdamBlockFilter* filter);


static GtkObjectClass* parent_class;

GtkType gdam_block_filter_get_type()
{
	static GtkType block_filter_type = 0;
	if (!block_filter_type) {
		static const GtkTypeInfo filter_info = {
			"GdamBlockFilter",
			sizeof(GdamBlockFilter),
			sizeof(GdamBlockFilterClass),
			(GtkClassInitFunc) gdam_block_filter_class_init,
			(GtkObjectInitFunc) gdam_block_filter_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		parent_class = gtk_type_class(GDAM_TYPE_FILTER);
		block_filter_type =
			gtk_type_unique(GDAM_TYPE_FILTER, &filter_info);
	}
	return block_filter_type;
}

static void gdam_block_filter_arg_set_iosizes(GtkObject* object, GdamArg* arg)
{
	GdamBlockFilter* block_filter = GDAM_BLOCK_FILTER(object);
	if (GDAM_ARG_INT_ARRAY(arg).length != 2) {
		gdam_debug("set_iosizes called with length of array != 2");
		return;
	}
	gdam_block_filter_set_sizes(block_filter,
                                    GDAM_ARG_INT_ARRAY(arg).array[0],
                                    GDAM_ARG_INT_ARRAY(arg).array[1]);
}
void gdam_block_filter_set_sizes(GdamBlockFilter* block_filter,
                                 int              input_size,
				 int              output_size)
{
	if (input_size != block_filter->input_block_size) {
		block_filter->input_block_buffer
				= g_renew(GdamSample,
				  	  block_filter->input_block_buffer,
					  input_size);
		if (block_filter->cur_input_length > input_size)
			block_filter->cur_input_length = input_size;
		block_filter->input_block_size = input_size;
	}
	if (output_size != block_filter->output_block_size) {
		block_filter->output_block_buffer
				= g_renew(GdamSample,
				  	  block_filter->output_block_buffer,
					  output_size);
		block_filter->output_block_size = output_size;
		if (block_filter->cur_output_length > output_size)
			block_filter->cur_output_length = output_size;
	}
}

static void gdam_block_filter_arg_get_iosizes(GtkObject* object, GdamArg* arg)
{
	GdamBlockFilter* block_filter = GDAM_BLOCK_FILTER(object);
	int* array = g_new(int, 2);
	GDAM_ARG_INT_ARRAY(arg).length = 2;
	GDAM_ARG_INT_ARRAY(arg).array = array;
	array[0] = block_filter->input_block_size;
	array[1] = block_filter->output_block_size;
}

static void gdam_block_filter_class_init(GdamBlockFilterClass* klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	GdamSourceClass* source_class = (GdamSourceClass*)klass;
	GdamFilterClass* filter_class = GDAM_FILTER_CLASS(klass);
	klass->process_block = gdam_block_filter_process_block;
	filter_class->attach_filtered = NULL;
	source_class->read = gdam_block_filter_read;
	source_class->miss_data = gdam_block_filter_miss_data;
	source_class->is_ready_to_read = gdam_block_filter_is_ready_to_read;
	object_class->destroy = (void(*)(GtkObject*))gdam_block_filter_destroy;
	gdam_arg_register(object_class,
	                  "input_output_sizes",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  gdam_block_filter_arg_set_iosizes,
			  gdam_block_filter_arg_get_iosizes,
			  GDAM_ARG_VITAL);
}

static void gdam_block_filter_init(GdamBlockFilter* filter)
{
	filter->input_block_size = 1024;
	filter->output_block_size = 1024;
	filter->output_block_buffer = g_new(GdamSample,
	                                    filter->output_block_size);
	filter->input_block_buffer = g_new(GdamSample,
	                                   filter->input_block_size);
}

static gboolean gdam_block_filter_is_ready_to_read(GdamSource* source)
{
	GdamBlockFilter* block_filter = GDAM_BLOCK_FILTER(source);
	GdamSource* filtered;
	if (block_filter->cur_output_length > 0)
		return TRUE;
	/* Hmm... we cannot really be sure, since there is no
	 * way to tell *how much* data will be read... sigh... */
	filtered = GDAM_FILTER(source)->filtered;
	return filtered ? gdam_source_is_ready_to_read(filtered) : TRUE;
}
static int 	gdam_block_filter_process_block    (GdamBlockFilter* filter,
	                                            GdamSample*      output,
					            const GdamSample*input)
{
	int input_size = filter->input_block_size;
	int output_size = filter->output_block_size;
	if (input_size < output_size) {
		gdam_sample_copy(output, input, input_size);
		gdam_sample_zero(output+input_size, output_size-input_size);
	} else {
		gdam_sample_copy(output, input, output_size);
	}
	return output_size;
}

static inline void do_block_transform(GdamBlockFilter* block_filter)
{
	GdamBlockFilterClass* block_class;
	int len;
	GtkObjectClass* object_class = ((GtkObject*)block_filter)->klass;
	block_class = GDAM_BLOCK_FILTER_CLASS(object_class);
	len = (*block_class->process_block)
				(block_filter,
				 block_filter->output_block_buffer,
				 block_filter->input_block_buffer);
	block_filter->cur_input_length = 0;
	block_filter->cur_output_length = len;
	block_filter->output_at = block_filter->output_block_buffer;
}

static int gdam_block_filter_one_read(GdamBlockFilter* block_filter,
                                      GdamSample* buf,
				      int n)
{
	GdamSource* source = (GdamSource*)block_filter;
	GdamFilter* filter = GDAM_FILTER(source);
	GdamSource* filtered = filter->filtered;
	int rv = 0;

	if (block_filter->cur_output_length > 0) {
		int ready = block_filter->cur_output_length;
		GdamSample* output_start;
		output_start = block_filter->output_block_buffer
		             + block_filter->output_block_size
			     - ready;
		if (ready > n) {
			gdam_sample_copy(buf, output_start, n);
			block_filter->cur_output_length -= n;
			block_filter->output_at += n;
			return rv + n;
		} else {
			gdam_sample_copy(buf, output_start, ready);
			rv = ready;
			block_filter->cur_output_length = 0;
		}
	}

	if (filtered == NULL)
		return rv;

	/* Read from underlying source. */
	if (block_filter->cur_input_length < block_filter->input_block_size
	 && gdam_source_is_ready_to_read(filtered)) {
		int cur_input = block_filter->cur_input_length;
		int num_read;
		GdamSample* input_end;
		input_end = block_filter->input_block_buffer + cur_input;
		num_read = block_filter->input_block_size - cur_input;
		num_read = gdam_source_read(filtered, input_end, num_read);
		if (num_read <= 0) {
			gdam_filter_set_source(filter, NULL);
			return rv;
		}
		block_filter->cur_input_length += num_read;
	}

	/* Do the block transformation. */
	if (block_filter->cur_input_length == block_filter->input_block_size
	 && block_filter->cur_output_length == 0) {
	 	do_block_transform(block_filter);
	}

	if (rv == 0 && block_filter->cur_output_length == 0) {
		/* Emergency block transform time. */
		if (block_filter->cur_input_length == 0) {
			gdam_sample_zero(block_filter->output_block_buffer,
			                 block_filter->output_block_size);
			block_filter->cur_output_length
					= block_filter->output_block_size;
			block_filter->output_at
					= block_filter->output_block_buffer;
		} else {
			do_block_transform(block_filter);
		}
	}
	if (block_filter->cur_output_length > 0) {
		int output_len = n - rv;
		if (output_len > block_filter->cur_output_length)
			output_len = block_filter->cur_output_length;
		gdam_sample_copy(buf + rv,
		                 block_filter->output_at,
				 output_len);
		rv += output_len;
		block_filter->cur_output_length -= output_len;
		block_filter->output_at += output_len;
	}
	return rv;
}

/* We loop around read_one until it has no more data to offer
 * us w/o blocking. */
static int gdam_block_filter_read(GdamSource* source, GdamSample* buf, int n)
{
	int num_read;
	int rv = 0;
	GdamBlockFilter* block_filter = GDAM_BLOCK_FILTER(source);
	gboolean got_data = FALSE;
	while ((num_read = gdam_block_filter_one_read(block_filter,
	                                              buf + rv,
						      n - rv)) > 0) {
		got_data = TRUE;
		if (block_filter->missed > 0) {
			int missed = block_filter->missed;
			if (missed > num_read) {
				missed -= num_read;
				num_read = 0;
			} else {
				num_read -= missed;
				memmove(buf,
				        buf + missed,
					sizeof(GdamSample) * num_read);
				missed = 0;
			}
			block_filter->missed = missed;
		}

		rv += num_read;
		if (rv == n) {
			return n;
		}
		if (!gdam_block_filter_is_ready_to_read(
						(GdamSource*)block_filter))
			break;
	}
	if (got_data && rv == 0) {
		/* KLUDGE: miss_data + not ready to read kludge */
		block_filter->missed++;
		buf->s[0] = buf->s[1] = 0;
		return 1;
	}
	return rv;
}

static void gdam_block_filter_miss_data(GdamSource* source, int lost)
{
	GDAM_BLOCK_FILTER(source)->missed += lost;
}


static void gdam_block_filter_destroy(GdamBlockFilter* block_filter)
{
	g_free(block_filter->output_block_buffer);
	g_free(block_filter->input_block_buffer);
	parent_class->destroy((GtkObject*)block_filter);
}
