/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamqueue.h"
#include "../gdamnamespace.h"

static GtkObjectClass* parent_class = NULL;

static void     gdam_queue_class_init       (GdamSourceClass*  klass);
static void     gdam_queue_init             (GdamQueue*        queue);
static void     gdam_queue_destroy          (GtkObject*        queue);
static GdamTime gdam_queue_get_position     (GdamSource*       input);
static int      gdam_queue_read             (GdamSource*       input,
                                             GdamSample*       samples,
		                             int               max_read);
static gboolean gdam_queue_seek             (GdamSource*      source,
                                             GdamTime         offset,
		                             int              whence);
static void     gdam_queue_get_subsources   (GdamSource*      source,
                                             int*             num_subs,
					     GdamSource***    subs);

static void     gdam_queue_miss_data        (GdamSource*       input,
                                             int               samples_lost);
static gboolean gdam_queue_is_ready_to_read (GdamSource* queue_source);
static gboolean gdam_queue_detach_child     (GdamSource*      mix,
                                             GdamSource*      sub_source);

static void gdam_queue_init(GdamQueue* queue)
{
	queue->terminates = 1;
}

static void gdam_queue_arg_enqueue(GtkObject* object, GdamArg* arg)
{
	int source_id = GDAM_ARG_INT(arg);
	GdamSource* source;
	GdamQueue* queue = GDAM_QUEUE(object);
	g_return_if_fail(source_id != 0);
	g_return_if_fail(queue != NULL);
	gdam_debug("en_queue: new num_subs=%d",GDAM_ARG_INT_ARRAY(arg).length);
	source = gdam_namespace_local_source(source_id);
	g_return_if_fail(source != NULL);
	gdam_queue_enqueue(queue, source);
}

static void gdam_queue_arg_get_queue(GtkObject* object, GdamArg* arg)
{
	int num_subsources;
	int* array;
	int i;
	GSList* at;
	GdamQueue* queue = GDAM_QUEUE(object);
	num_subsources = g_slist_length(queue->first_source);
	array = g_new(int, num_subsources);
	at = queue->first_source;
	for (i = 0; i < num_subsources; i++) {
		array[i] = GDAM_SOURCE(at->data)->source_id;
		at = at->next;
	}
	GDAM_ARG_INT_ARRAY(arg).length = num_subsources;
	GDAM_ARG_INT_ARRAY(arg).array = array;
}
static void reparent(gpointer source, gpointer queue)
{
	GdamSource* child = GDAM_SOURCE(source);
	GdamSource* parent = GDAM_SOURCE(queue);
	if (child->parent != NULL)
		gdam_source_detach_child(child->parent, child);
	child->parent = parent;
}

static void gdam_queue_arg_set_queue(GtkObject* object, GdamArg* arg)
{
	GSList* new_queue = NULL;
	GdamQueue* queue = GDAM_QUEUE(object);
	int i;
	int length = GDAM_ARG_INT_ARRAY(arg).length;
	int* array = GDAM_ARG_INT_ARRAY(arg).array;
	for (i = 0; i < length; i++) {
		int id = array[i];
		GdamSource* new_source = gdam_namespace_local_source(id);
		GtkObject* new_object;
		if (new_source == NULL) {
			gdam_debug("queue: no source of id %d found", id);
			continue;
		}
		new_object = GTK_OBJECT(new_source);
		gtk_object_ref(new_object);
		gtk_object_sink(new_object);
		new_queue = g_slist_prepend(new_queue, new_object);
	}
	while (queue->first_source)
		gdam_queue_dequeue(queue);
	queue->first_source = g_slist_reverse(new_queue);
	queue->last_source = g_slist_last(queue->first_source);
	g_slist_foreach(queue->first_source, reparent, queue);
}

static void gdam_queue_arg_set_terminates(GtkObject* object, GdamArg* arg)
{
	GDAM_QUEUE(object)->terminates = GDAM_ARG_INT(arg) ? 1 : 0;
}
static void gdam_queue_arg_get_terminates(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_QUEUE(object)->terminates;
}


static void gdam_queue_class_init(GdamSourceClass* source_class)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(source_class);
	object_class->destroy = gdam_queue_destroy;
	source_class->read = gdam_queue_read;
	source_class->miss_data = gdam_queue_miss_data;
	source_class->is_ready_to_read = gdam_queue_is_ready_to_read;
	source_class->detach_child = gdam_queue_detach_child;
	source_class->get_position = gdam_queue_get_position;
	source_class->get_subsources = gdam_queue_get_subsources;
	source_class->seek = gdam_queue_seek;
	gdam_arg_register(object_class,
	                  "enqueue",
			  GDAM_ARG_TYPE_INT,
			  gdam_queue_arg_enqueue,
			  NULL,
			  0);
	gdam_arg_register(object_class,
	                  "queue",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  gdam_queue_arg_set_queue,
			  gdam_queue_arg_get_queue,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "terminates",
			  GDAM_ARG_TYPE_INT,
			  gdam_queue_arg_set_terminates,
			  gdam_queue_arg_get_terminates,
			  GDAM_ARG_VITAL);
}

GtkType gdam_queue_get_type()
{
	static GtkType queue_type = 0;
	if (!queue_type) {
		static const GtkTypeInfo queue_info = {
			"GdamQueue",
			sizeof(GdamQueue),
			sizeof(GdamQueueClass),
			(GtkClassInitFunc) gdam_queue_class_init,
			(GtkObjectInitFunc) gdam_queue_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SOURCE;
		parent_class = gtk_type_class(parent);
		queue_type = gtk_type_unique(parent, &queue_info);
	}
	return queue_type;
}

static void     gdam_queue_destroy          (GtkObject*        object)
{
	GdamQueue* queue = GDAM_QUEUE(object);
	while (queue->first_source)
		gdam_queue_dequeue(queue);
	(*parent_class->destroy)(object);
}


static int      gdam_queue_read             (GdamSource*       input,
                                             GdamSample*       samples,
		                             int               max_read)
{
	GdamQueue* queue = GDAM_QUEUE(input);
	int rv = 0;
	while (queue->first_source) {
		GdamSource* subsource;
		int num_read;
		subsource = queue->first_source->data;
		if (!gdam_source_is_ready_to_read(subsource)) {
			/* XXX: fix this typical hack... */
			if (rv == 0) {
				gdam_sample_zero(samples, 1);
				rv++;
			}
			return rv;
		}
		num_read = gdam_source_read(subsource,
		                            samples + rv,
					    max_read - rv);
		if (num_read == 0) {
			gdam_queue_dequeue(queue);
			continue;
		}
		if (num_read == max_read - rv)
			return max_read;

		rv += num_read;
	}
	if (queue->first_source == NULL) {
		if (queue->terminates)
			return rv;
		gdam_sample_zero(samples + rv, max_read - rv);
		return max_read;
	}
	return rv;
}

static void     gdam_queue_miss_data        (GdamSource*       input,
                                             int               samples_lost)
{
	GdamQueue* queue = GDAM_QUEUE(input);
	GdamSource* subsource;
	if (queue == NULL || queue->first_source == NULL)
		return;
	subsource = GDAM_SOURCE(queue->first_source->data);
	gdam_source_miss_data(subsource, samples_lost);
}

static gboolean gdam_queue_is_ready_to_read (GdamSource* queue_source)
{
	GdamQueue* queue = GDAM_QUEUE(queue_source);
	GdamSource* first;
	g_return_val_if_fail(queue != NULL, FALSE);
	if (queue->first_source == NULL)
		return TRUE;
	first = GDAM_SOURCE(queue->first_source->data);
	return gdam_source_is_ready_to_read(first);
}

static gboolean gdam_queue_detach_child     (GdamSource*      source,
                                             GdamSource*      sub_source)
{
	GdamQueue* queue = GDAM_QUEUE(source);
	GSList* prev = NULL;
	GSList* at;
	for (at = queue->first_source; at != NULL; at = at->next) {
		if (at->data == sub_source)
			break;
		prev = at;
		at = at->next;
	}
	if (at == NULL)
		return FALSE;
	if (prev)
		prev->next = at->next;
	else
		queue->first_source = at->next;
	if (at == queue->last_source)
		queue->last_source = prev;
	g_slist_free_1(at);
	sub_source->parent = NULL;
	gtk_object_unref(GTK_OBJECT(sub_source));
	return TRUE;
}

GdamQueue* gdam_queue_new()
{
	GdamQueue* queue = GDAM_QUEUE(gtk_type_new(GDAM_TYPE_QUEUE));
	return queue;
}

void gdam_queue_enqueue(GdamQueue*    queue,
                        GdamSource*   source)
{
	GtkObject* sub_object = GTK_OBJECT(source);
	if (source->parent != NULL) {
		gdam_source_detach_child(source->parent, source);
	}
	source->parent = GDAM_SOURCE(queue);
	if (queue->first_source == NULL)
		queue->first_source = queue->last_source
				    = g_slist_prepend(NULL, source);
	else {
		g_slist_append(queue->last_source, source);
		queue->last_source = queue->last_source->next;
	}
	g_return_if_fail(queue->first_source != NULL);
	g_return_if_fail(queue->last_source != NULL);
	gtk_object_ref(sub_object);
	gtk_object_sink(sub_object);
}

void gdam_queue_dequeue(GdamQueue*    queue)
{
	GSList* first = queue->first_source;
	GSList* next;
	GdamSource* subsource;
	g_return_if_fail(first != NULL);
	next = first->next;

	subsource = GDAM_SOURCE(queue->first_source->data);
	g_slist_free_1(queue->first_source);
	queue->first_source = next;
	if (queue->first_source == NULL)
		queue->last_source = NULL;
	subsource->parent = NULL;
	gtk_object_unref(GTK_OBJECT(subsource));
}
static gboolean gdam_queue_seek             (GdamSource*      source,
                                             GdamTime         offset,
		                             int              whence)
{
	GdamQueue* queue = GDAM_QUEUE(source);
	GSList* first = queue->first_source;
	GdamSource* subsource;
	if (first == NULL)
		return FALSE;
	subsource = GDAM_SOURCE(first->data);
	return gdam_source_seek(subsource, offset, whence);
}

static GdamTime gdam_queue_get_position     (GdamSource*      source)
{
	GdamQueue* queue = GDAM_QUEUE(source);
	GSList* first = queue->first_source;
	GdamSource* subsource;
	if (first == NULL)
		return -1;
	subsource = GDAM_SOURCE(first->data);
	return gdam_source_get_position(subsource);
}
static void     gdam_queue_get_subsources   (GdamSource*      source,
                                             int*             num_subs,
					     GdamSource***    subs)
{
	GdamQueue* queue = GDAM_QUEUE(source);
	GSList* at = queue->first_source;
	int i;
	*num_subs = g_slist_length(at);
	*subs = g_new(GdamSource*, *num_subs);
	for (i = 0; at != NULL; at = at->next, i++)
		(*subs)[i] = at->data;
}
